@@ -2,6 +2,7 @@
 #include "pattern2.h"
 #include "incubator.h"
 #include <unordered_map>
+#include <map>
 #include <set>
 
 #include "dsds.h"
@@ -39,468 +40,351 @@ namespace apg {
             return (transtable[257] || transtable[68] || b0);
         }
 
+        std::vector<std::vector<apg::pattern>> get_patterns(dsds &uf, std::unordered_map<uint64_t, uint64_t> &umap, std::vector<apg::pattern> &phases, std::vector<int> *short_connections) {
+
+            uint64_t islcount = 0;
+
+            std::unordered_map<uint64_t, uint64_t> cmap;
+            std::vector<uint64_t> invcmap;
+
+            for (auto it = umap.begin(); it != umap.end(); ++it) {
+                uint64_t cid = uf.find(it->second);
+                if (cmap.count(cid) == 0) {
+                    cmap[cid] = (islcount++);
+                    invcmap.push_back(cid);
+                }
+            }
+
+            // Create empty array of patterns:
+            std::vector<std::vector<apg::pattern>> subpats(islcount);
+            for (uint64_t c = 0; c < islcount; c++) {
+                for (uint64_t t = 0; t < phases.size(); t++) {
+                    subpats[c].emplace_back(lab, "", rule);
+                }
+            }
+
+            // Populate patterns:
+            for (auto it = umap.begin(); it != umap.end(); ++it) {
+                uint64_t cid = uf.find(it->second);
+                uint64_t c = cmap[cid];
+                uint64_t t = it->first >> 48;
+                subpats[c][t] += phases[t].subrect((it->first & 0xffffff), ((it->first >> 24) & 0xffffff), 1, 1);
+            }
+
+            if (short_connections == nullptr) { return subpats; }
+
+            for (uint64_t t = 0; t < phases.size() - 1; t++) {
+                for (uint64_t i = 0; i < islcount; i++) {
+                    auto successor = subpats[i][t][1];
+                    if (successor == subpats[i][t+1]) {
+                        continue; // island is stable in this generation
+                    }
+                    auto extrabirths = successor - subpats[i][t+1];
+                    std::vector<int64_t> celllist(extrabirths.totalPopulation() * 2);
+                    extrabirths.get_coords(&(celllist[0]));
+
+                    for (uint64_t ii = 0; ii < celllist.size(); ii += 2) {
+                        uint64_t coord = (t << 48) + (celllist[ii+1] << 24) + celllist[ii];
+                        
+                        std::set<uint64_t> friends;
+
+                        for (auto&& x : (*short_connections)) {
+                            uint64_t other = coord + x;
+                            if (umap.count(other)) {
+                                uint64_t j = cmap[uf.find(umap[other])];
+                                if (j != i) { friends.insert(j); }
+                            }
+                        }
+
+                        if (friends.size() == 1) {
+                            uint64_t j = *(friends.begin());
+                            uf.merge(invcmap[i], invcmap[j]);
+                            return subpats; // improvement made
+                        } else if (friends.size() == 0) {
+                            std::cerr << "Unreachable code!!!" << std::endl;
+                        }
+                    }
+                }
+            }
+
+            return subpats;
+        }
+
         /**
-         * Better object separation, especially for spaceships:
+         * Complete object separation.
          */
-        std::vector<std::string> map_3dsep(const std::vector<std::string> &apgcodes) {
+        std::vector<std::string> full_3dsep(std::string parent) {
+
+            // cache:
+            {
+                auto it = decompositions.find(parent);
+                if (it != decompositions.end()) {
+                    return it->second;
+                }
+            }
 
             std::vector<std::string> elements;
 
-            for (auto&& parent : apgcodes) {
+            #define RETURN_SINGLETON() elements.push_back(parent); decompositions[parent] = elements; return elements
+            if (parent[0] != 'x') { RETURN_SINGLETON(); }
 
-                if (parent[0] != 'x') {
-                    elements.push_back(parent);
-                    continue;
+            apg::pattern pat(lab, parent, rule);
+            uint64_t period = pat.ascertain_period();
+
+            if ((pat.empty()) || (period == 0) || (period > 30000)) { RETURN_SINGLETON(); }
+
+            int64_t dy = pat.dy;
+            int64_t dx = pat.dx;
+
+            pat = pat.shift(262144, 262144);
+
+            int64_t bbox[4] = {0};
+            pat.getrect(bbox);
+
+            if ((bbox[2] > 100000) || (bbox[3] > 100000)) { RETURN_SINGLETON(); }
+
+            std::unordered_map<uint64_t, uint64_t> umap;
+            uint64_t id = 0;
+
+            std::vector<apg::pattern> phases;
+
+            for (uint64_t t = 0; t <= period; t++) {
+                phases.emplace_back(lab, "", rule);
+                phases[t] += pat;
+
+                std::vector<int64_t> celllist(pat.totalPopulation() * 2);
+                pat.get_coords(&(celllist[0]));
+                for (uint64_t i = 0; i < celllist.size(); i += 2) {
+                    uint64_t coord = (t << 48) + (celllist[i+1] << 24) + celllist[i];
+                    umap[coord] = id; id += 1;
                 }
 
-                apg::pattern pat(lab, parent, rule);
-                uint64_t period = pat.ascertain_period();
+                if (t != period) { pat = pat[1]; }
+            }
 
-                if ((pat.empty()) || (period == 0) || (period > 30000)) {
-                    elements.push_back(parent);
-                    continue;
+            std::vector<uint64_t> connections;
+            std::vector<int> short_connections;
+
+            connections.push_back((period << 48) + (dy << 24) + dx);
+
+            int range = zoi.size() >> 1;
+
+            for (int i = -range; i <= range; i++) {
+                for (int j = -range; j <= range; j++) {
+                    int k = (i << 24) + j;
+                    if (k > 0) { connections.push_back(k); }
+                    connections.push_back((1ull << 48) + k);
+                    if (k != 0) { short_connections.push_back(k); }
                 }
+            }
 
-                pat = pat.shift(262144, 262144);
+            // disjoint-set data structure:
+            dsds uf(id);
+            dsds uf2(id);
 
-                int64_t bbox[4] = {0};
-                pat.getrect(bbox);
+            std::map<uint64_t, std::vector<uint64_t>> suppressed;
 
-                if ((bbox[2] > 100000) || (bbox[3] > 100000)) {
-                    elements.push_back(parent);
-                    continue;
-                }
-
-                std::unordered_map<uint64_t, uint64_t> umap;
-                uint64_t id = 0;
-
-                {
-                    apg::pattern pat2 = pat;
-
-                    for (uint64_t t = 0; t <= period; t++) {
-                        std::vector<int64_t> celllist(pat2.totalPopulation() * 2);
-                        pat2.get_coords(&(celllist[0]));
-                        for (uint64_t i = 0; i < celllist.size(); i += 2) {
-                            uint64_t coord = (t << 48) + (celllist[i+1] << 24) + celllist[i];
-                            umap[coord] = id; id += 1;
-                        }
-                        pat2 = pat2[1];
+            for (auto it = umap.begin(); it != umap.end(); ++it) {
+                for (auto&& x : connections) {
+                    uint64_t other = it->first + x;
+                    if (umap.count(other)) {
+                        uf.merge(it->second, umap[other]);
+                        uf2.merge(it->second, umap[other]);
                     }
                 }
-
-                std::vector<uint64_t> connections;
-
-                connections.push_back((period << 48) + (pat.dy << 24) + pat.dx);
-
-                int range = zoi.size() >> 1;
-
-                for (int i = 0; i <= range*2; i++) {
-                    for (int j = -range*2; j <= range*2; j++) {
-                        int k = (i << 24) + j;
-                        if (k > 0) { connections.push_back(k); }
-                    }
-                }
-
-                for (int i = -range; i <= range; i++) {
-                    for (int j = -range; j <= range; j++) {
-                        connections.push_back((1ull << 48) + (i << 24) + j);
-                    }
-                }
-
-                // disjoint-set data structure:
-                dsds uf(id);
-
-                for (auto it = umap.begin(); it != umap.end(); ++it) {
-                    for (auto&& x : connections) {
-                        uint64_t other = it->first + x;
-                        if (umap.count(other)) { uf.merge(it->second, umap[other]); }
-                    }
-                }
-
-                {
-                    std::vector<apg::pattern> subpats;
-                    std::unordered_map<uint64_t, uint64_t> cmap;
-
-                    for (auto it = umap.begin(); it != umap.end(); ++it) {
-                        if (it->first >= (1ull << 48)) { continue; }
-
-                        uint64_t cid = uf.find(it->second);
-                        if (cmap.count(cid) == 0) {
-                            cmap[cid] = subpats.size();
-                            subpats.emplace_back(lab, "", rule);
-                        }
-
-                        subpats[cmap[cid]] += pat.subrect((it->first & 0xffffff), (it->first >> 24), 1, 1);
-                    }
-
-                    for (auto&& spat : subpats) {
-                        elements.push_back(spat.apgcode());
+                for (auto&& x : short_connections) {
+                    uint64_t other = it->first + x;
+                    if (!(umap.count(other))) {
+                        suppressed[other].push_back(it->second);
                     }
                 }
             }
 
-            return elements;
-        }
+            for (auto it = suppressed.begin(); it != suppressed.end(); ++it) {
+                uint64_t l = it->second.size();
+                if (l >= 3) {
+                    for (uint64_t i = 1; i < l; i++) {
+                        uf2.merge(it->second[i-1], it->second[i]);
+                    }
+                }
+            }
+
+            if (uf2.count() > 1) {
+                // fast separation:
+
+                auto subpats = get_patterns(uf2, umap, phases, nullptr);
+
+                for (auto&& y : subpats) {
+                    std::string apgcode = y[0].apgcode();
+                    auto new_elems = full_3dsep(apgcode);
+                    for (auto&& x : new_elems) { elements.push_back(x); }
+                }
+
+                decompositions[parent] = elements;
+                return elements;
+            }
+
+            std::vector<std::vector<apg::pattern>> subpats;
+
+            // uint64_t isize = uf.count();
+
+            while (subpats.size() != uf.count()) {
+                subpats = get_patterns(uf, umap, phases, &short_connections);
+            }
+
+            uint64_t islcount = subpats.size();
 
-        std::vector<std::string> pbbosc(pattern pat, uint64_t n, uint64_t maxn) {
             /*
-            * Exhaustively search all partitions of an oscillator into
-            * n disjoint unions of islands to check whether this is a
-            * pseudo-oscillator.
+            if (islcount != isize) {
+                std::cout << parent << " improved from " << isize << " islands to " << islcount << std::endl;
+            }
             */
 
-            std::vector<std::string> apgcodes;
+            // Only one island; terminate.
+            if (islcount == 1) { RETURN_SINGLETON(); }
 
-            uint64_t period = pat.ascertain_period();
-            pattern hist(&lh, "", rule + "History");
-            hist += pat;
-            hist = hist[period + 2];
-            bitworld lrem = hist.flatlayer(0);
-            bitworld env = hist.flatlayer(1);
-
-            apg::pattern clunion(lab, lab->demorton(lrem, 1), rule);
-
-            std::vector<pattern> clusters;
-            while (lrem.population() != 0) {
-                bitworld cluster = grow_cluster(lrem.get1cell(), env, "9");
-                lrem -= cluster;
-                apg::pattern ppart(lab, lab->demorton(cluster, 1), rule);
-                clusters.push_back(ppart & clunion);
+            bool all_standalone = true;
+            for (uint64_t i = 0; i < islcount; i++) {
+                for (uint64_t t = 0; t < period; t++) {
+                    if (subpats[i][t][1] != subpats[i][t+1]) {
+                        all_standalone = false;
+                    }
+                }
             }
 
-            uint64_t islcount = clusters.size();
+            if (all_standalone) {
 
-            if (islcount < n) { apgcodes.push_back(pat.apgcode()); return apgcodes; }
+                for (uint64_t i = 0; i < islcount; i++) {
+                    std::string apgcode = subpats[i][0].apgcode();
+                    auto new_elems = full_3dsep(apgcode);
+                    for (auto&& x : new_elems) { elements.push_back(x); }
+                }
 
-            uint64_t bftime = modexp_u64(n, islcount, -1);
-            if (((period + 2) * bftime) > 1000000) {
-                // std::cerr << pat.apgcode() << " would take infeasibly long to brute-force separate." << std::endl;
-                apgcodes.push_back(pat.apgcode());
-                return apgcodes;
+                decompositions[parent] = elements;
+                return elements;
             }
 
-            std::vector<uint8_t> currstack(islcount);
-            std::vector<uint8_t> maxstack(islcount);
+            std::set<uint64_t> invalid_cache;
 
-            currstack[0] = 0;
-            currstack[1] = 0;
-            maxstack[0] = 0;
+            bool low_period_osc = (period <= 3) && (dx == 0) && (dy == 0);
 
-            uint64_t focus = 1;
-            while (focus) {
+            for (uint64_t n = 2; n <= islcount; n++) {
 
-                uint8_t limit = maxstack[focus - 1] + 1;
-                limit = (limit >= n) ? (n - 1) : limit;
-                // for (uint64_t i = 0; i <= focus; i++) {
-                //     std::cerr << " " << ((int) currstack[i]);
-                // }
-                // std::cerr << std::endl;
-                if (currstack[focus] > limit) {
-                    focus -= 1;
-                    currstack[focus] += 1;
-                } else {
+                if (low_period_osc) {
+                    if ((n > 4) && (zoi == "95")) {
+                        // unnecessary by Four Colour Theorem
+                        break;
+                    } else if ((n > 7) && (zoi == "99")) {
+                        break;
+                    }
+                }
 
-                    maxstack[focus] = maxstack[focus - 1];
-                    if (maxstack[focus] < currstack[focus]) { maxstack[focus] = currstack[focus]; }
-                    if (focus < (islcount - 1)) {
-                        focus += 1;
-                        currstack[focus] = 0;
-                    } else {
-                        if (maxstack[focus] == n - 1) {
-                            // We have a n-colouring which uses all n colours:
-                            std::vector<pattern> unions;
-                            for (uint64_t i = 0; i < n; i++) { unions.push_back(pattern(lab, "", rule)); }
-                            for (uint64_t i = 0; i < islcount; i++) {
-                                unions[currstack[i]] |= clusters[i];
-                            }
-                            bool faithful = true;
-                            for (uint64_t i = 0; i <= period; i++) {
-                                apg::pattern x = unions[0];
-                                unions[0] = unions[0][1];
-                                if ((unions[0]).empty()) { faithful = false; break; }
-                                for (uint64_t j = 1; j < n; j++) {
-                                    if ((unions[j]).empty()) { faithful = false; break; }
-                                    if ((x & unions[j]).nonempty()) { faithful = false; break; }
-                                    x |= unions[j];
-                                    unions[j] = unions[j][1];
-                                }
-                                if (x != clunion[i]) { faithful = false; break; }
-                                if (faithful == false) { break; }
-                            }
-                            if (faithful) {
-                                // We have a decomposition into non-interacting pieces!
-                                for (uint64_t i = 0; i < n; i++) {
-                                    std::vector<std::string> tx = pbbosc(unions[i], 2, maxn);
-                                    for (uint64_t j = 0; j < tx.size(); j++) {
-                                        apgcodes.push_back(tx[j]);
-                                    }
-                                }
-                                return apgcodes;
-                            }
-                        }
+                uint64_t cost = 1;
+                for (uint64_t i = 0; i < islcount; i++) {
+                    cost *= n;
+                    if (cost > 1048576) { break; }
+                }
+
+                if ((cost > 1048576) || (islcount > 16)) {
+                    // too expensive; give up
+                    std::cerr << "Gave up attempting to separate " << parent << " (" << islcount << " islands) into " << n << " subobjects." << std::endl;
+                    break;
+                }
+
+                // std::cerr << "Separating " << parent << " (" << islcount << " islands) into " << n << " subobjects." << std::endl;
+
+                std::vector<uint8_t> currstack(islcount);
+                std::vector<uint8_t> maxstack(islcount);
+
+                currstack[0] = 0;
+                currstack[1] = 0;
+                maxstack[0] = 0;
+
+                uint64_t focus = 1;
+                while (focus) {
+
+                    uint8_t limit = maxstack[focus - 1] + 1;
+                    limit = (limit >= n) ? (n - 1) : limit;
+                    if (currstack[focus] > limit) {
+                        focus -= 1;
                         currstack[focus] += 1;
-                    }
-                }
-            }
+                    } else {
 
-            // std::cerr << " -- reached end without partition" << std::endl;
+                        maxstack[focus] = maxstack[focus - 1];
+                        if (maxstack[focus] < currstack[focus]) { maxstack[focus] = currstack[focus]; }
+                        if (focus < (islcount - 1)) {
+                            focus += 1;
+                            currstack[focus] = 0;
+                        } else {
+                            if (maxstack[focus] == n - 1) {
 
-            if ((maxn == 0) || (n < maxn)) {
-                // Try with more partitions:
-                return pbbosc(pat, n + 1, maxn);
-            } else {
-                // Pattern cannot be decomposed:
-                apgcodes.push_back(pat.apgcode());
-                return apgcodes;
-            }
-        }
+                                // We have a n-colouring which uses all n colours:
+                                std::vector<uint64_t> bitmasks(n);
 
-        std::vector<std::string> pseudoBangBang(pattern pat, std::vector<bitworld> *clvec, bool recurse) {
-            /*
-            * Borrowed from apgmera, and upgraded.
-            */
-
-            uint64_t period = pat.ascertain_period();
-            bool isOscillator = ((pat.dx == 0) && (pat.dy == 0));
-            pattern hist(&lh, "", rule + "History");
-            hist += pat;
-            hist = hist[period + 2];
-            bitworld lrem = hist.flatlayer(0);
-            bitworld env = hist.flatlayer(1);
-
-            // If we have a moving object, do not reiterate:
-            bool reiterate = isOscillator && (zoi.length() <= 2);
-
-            std::map<std::pair<int64_t, int64_t>, uint64_t> geography;
-            uint64_t label = 0;
-            while (lrem.population() != 0) {
-                bitworld cluster = grow_cluster(lrem.get1cell(), env, reiterate ? "9" : zoi);
-                lrem -= cluster;
-                label += 1;
-                std::vector<std::pair<int64_t, int64_t> > celllist = cluster.getcells();
-                for (uint64_t i = 0; i < celllist.size(); i++) {
-                    geography[celllist[i]] = label;
-                }
-            }
-
-            while (reiterate) {
-                reiterate = false;
-                for (uint64_t i = 0; i < period; i++) {
-                    hist = hist[1];
-                    bitworld lcurr = hist.flatlayer(0);
-                    bitworld dcurr = bleed(lcurr, "9");
-                    dcurr -= env;
-                    std::vector<std::pair<int64_t, int64_t> > liberties = dcurr.getcells();
-                    for (uint64_t j = 0; j < liberties.size(); j++) {
-                        int64_t ix = liberties[j].first;
-                        int64_t iy = liberties[j].second;
-                        std::map<uint64_t, uint64_t> tally;
-                        for (int64_t ux = 0; ux <= 2; ux++) {
-                            for (int64_t uy = 0; uy <= 2; uy++) {
-                                int value = geography[std::pair<int64_t, int64_t>(ux + ix - 1, uy + iy - 1)];
-                                if (lcurr.getcell(ux + ix - 1, uy + iy - 1)) {
-                                    tally[value] = tally[value] + (1 << (uy * 3 + ux));
+                                for (uint64_t i = 0; i < islcount; i++) {
+                                    bitmasks[currstack[i]] |= (1ull << i);
                                 }
-                            }
-                        }
 
-                        uint64_t dominantColour = 0;
-                        std::map<uint64_t, uint64_t>::iterator it2;
-                        for (it2 = tally.begin(); it2 != tally.end(); it2++) {
-                            int colour = it2->first;
-                            uint64_t count = it2->second;
-                            if (transtable[count]) { dominantColour = colour; }
-                            // if (__builtin_popcountll(count) == 3) { dominantColour = colour; }
-                        }
-                        // Resolve dependencies:
-                        if (dominantColour != 0) {
-                            std::map<std::pair<int64_t, int64_t>, uint64_t>::iterator it3;
-                            for (it3 = geography.begin(); it3 != geography.end(); it3++) {
-                                std::pair<int64_t, int64_t> coords = it3->first;
-                                uint64_t colour = it3->second;
+                                bool valid = true;
 
-                                if (tally[colour] > 0) {
-                                    geography[coords] = dominantColour;
-                                    if (colour != dominantColour) {
-                                        // A change has occurred; keep iterating until we achieve stability:
-                                        reiterate = true;
+                                for (uint64_t i = 0; i < n; i++) {
+                                    if (invalid_cache.count(bitmasks[i])) { valid = false; }
+                                }
+
+                                if (valid) {
+
+                                    std::vector<std::vector<apg::pattern>> unions(n);
+                                    for (uint64_t i = 0; i < n; i++) {
+                                        for (uint64_t t = 0; t <= period; t++) {
+                                            unions[i].push_back(pattern(lab, "", rule));
+                                        }
+                                    }
+                                    for (uint64_t i = 0; i < islcount; i++) {
+                                        for (uint64_t t = 0; t <= period; t++) {
+                                            unions[currstack[i]][t] |= subpats[i][t];
+                                        }
+                                    }
+
+                                    bool faithful = true;
+
+                                    for (uint64_t i = 0; i < n; i++) {
+                                        for (uint64_t t = 0; t < period; t++) {
+                                            if (unions[i][t][1] != unions[i][t+1]) {
+                                                faithful = false;
+                                                invalid_cache.insert(bitmasks[i]);
+                                            }
+                                        }
+                                    }
+
+                                    if (faithful) {
+                                        // We have a decomposition into non-interacting pieces!
+
+                                        for (uint64_t i = 0; i < n; i++) {
+                                            std::string apgcode = unions[i][0].apgcode();
+                                            auto new_elems = full_3dsep(apgcode);
+                                            for (auto&& x : new_elems) { elements.push_back(x); }
+                                        }
+
+                                        decompositions[parent] = elements;
+                                        return elements;
                                     }
                                 }
                             }
+                            currstack[focus] += 1;
                         }
                     }
                 }
             }
 
-            bitworld lcurr = (isOscillator ? pat.flatlayer(0) : hist.flatlayer(0));
-            std::vector<bitworld> cbs(label+1);
-            std::map<std::pair<int64_t, int64_t>, uint64_t>::iterator it3;
-            for (it3 = geography.begin(); it3 != geography.end(); it3++) {
-                std::pair<int64_t, int64_t> coords = it3->first;
-                uint64_t colour = it3->second;
-                cbs[colour].setcell(coords.first, coords.second, 1);
-            }
-            std::vector<std::string> components;
-            for (uint64_t l = 1; l <= label; l++) {
-                cbs[l] &= lcurr;
-                if (cbs[l].population() > 0) {
-                    if (clvec != 0) { clvec->push_back(cbs[l]); }
-                    apg::pattern ppart(lab, lab->demorton(cbs[l], 1), rule);
-                    if (isOscillator) {
-                        /*
-                        * Without diagbirths, we only need to go up to 4 as a
-                        * result of the Four Colour Theorem:
-                        *
-                        *   http://www.paradise.caltech.edu/~cook/Workshop/CAs/2DOutTot/Life/StillLife/FourColors.html
-                        *
-                        * With diagbirths, 7 may be necessary:
-                        *
-                        *   http://link.springer.com/article/10.1007/s10958-014-1693-6
-                        *
-                        * Either way, we start with 2 and increment:
-                        */
-                        uint64_t maxn = (zoi.length() <= 2) ? (diagbirth() ? 7 : 4) : 0;
-                        std::vector<std::string> tx = pbbosc(ppart, 2, maxn);
-                        for (uint64_t j = 0; j < tx.size(); j++) {
-                            components.push_back(tx[j]);
-                        }
-                        // std::cerr << "done!" << std::endl;
-                    } else if (recurse) {
-                        auto cmap = census(ppart, ppart.ascertain_period(), 0, false);
-                        if (cmap.size() == 0) {
-                            std::cerr << "Warning: " << ppart.apgcode() << " gave empty census." << std::endl;
-                            components.push_back(ppart.apgcode());
-                        }
-                        for (auto it = cmap.begin(); it != cmap.end(); ++it) {
-                            for (int i = 0; i < it->second; i++) {
-                                components.push_back(it->first);
-                            }
-                        }
-                    } else {
-                        components.push_back(ppart.apgcode());
-                    }
-                }
-            }
-
-            for (auto it = components.begin(); it != components.end(); ++it) {
-                if ((*it) == "PATHOLOGICAL") {
-                    // invalid separation
-                    std::vector<std::string> newcomps;
-                    newcomps.push_back(pat.apgcode());
-                    return newcomps;
-                }
-            }
-
-            return components;
-        }
-
-        /*
-        * Compute the degree sequence of the graph where vertices are live
-        * cells and edges denote adjacency. Concatenate the resulting
-        * sequences for each generation.
-        */
-        void degcount(pattern pat, int *degrees, int generations) {
-
-            pattern x = pat;
-            for (int i = 0; i < generations; i++) {
-                for (int j = 0; j < 9; j++) { degrees[9*i + j] = 0; }
-                bitworld bw = x.flatlayer(0);
-                x = x[1];
-
-                std::vector<std::pair<int64_t, int64_t> > celllist = bw.getcells();
-                std::set<std::pair<int64_t, int64_t> > cellset;
-                for (uint64_t k = 0; k < celllist.size(); k++) {
-                    cellset.insert(celllist[k]);
-                }
-                for (uint64_t k = 0; k < celllist.size(); k++) {
-                    int64_t x = celllist[k].first;
-                    int64_t y = celllist[k].second;
-
-                    int degree = -1;
-
-                    for (int64_t ix = x - 1; ix <= x + 1; ix++) {
-                        for (int64_t iy = y - 1; iy <= y + 1; iy++) {
-                            degree += cellset.count(std::pair<int64_t, int64_t>(ix, iy));
-                        }
-                    }
-
-                    degrees[9*i + degree] += 1;
-                }
-            }
-        }
-
-        /*
-        * Separate a collection of period-4 standard spaceships:
-        */
-        std::vector<std::string> sss(apg::pattern pat) {
-
-            std::vector<std::string> components;
-
-            int degrees[36];
-            degcount(pat, degrees, 4);
-
-            for (int i = 0; i < 18; i++) {
-                if (degrees[i] != degrees[18+i]) { return components; }
-            }
-
-            int hwssa[18] = {1,4,6,2,0,0,0,0,0,0,0,0,4,4,6,1,2,1};
-            int mwssa[18] = {2,2,5,2,0,0,0,0,0,0,0,0,4,4,4,1,2,0};
-            int lwssa[18] = {1,2,4,2,0,0,0,0,0,0,0,0,4,4,2,2,0,0};
-            int hwssb[18] = {0,0,0,4,4,6,1,2,1,1,4,6,2,0,0,0,0,0};
-            int mwssb[18] = {0,0,0,4,4,4,1,2,0,2,2,5,2,0,0,0,0,0};
-            int lwssb[18] = {0,0,0,4,4,2,2,0,0,1,2,4,2,0,0,0,0,0};
-            int glida[18] = {0,1,2,1,1,0,0,0,0,0,2,1,2,0,0,0,0,0};
-            int glidb[18] = {0,2,1,2,0,0,0,0,0,0,1,2,1,1,0,0,0,0};
-
-            int hacount = degrees[17];
-            int macount = degrees[16]/2 - hacount;
-            int lacount = (degrees[15] - hacount - macount)/2;
-            int hbcount = degrees[8];
-            int mbcount = degrees[7]/2 - hbcount;
-            int lbcount = (degrees[6] - hbcount - mbcount)/2;
-
-            int gacount = 0;
-            int gbcount = 0;
-
-            if ((lacount == 0) && (lbcount == 0) && (macount == 0) && (mbcount == 0) && (hacount == 0) && (hbcount == 0)) {
-                gacount = degrees[4];
-                gbcount = degrees[13];
-            }
-
-            for (int i = 0; i < 18; i++) {
-                int putativedegrees = 0;
-                putativedegrees += hacount * hwssa[i];
-                putativedegrees += hbcount * hwssb[i];
-                putativedegrees += lacount * lwssa[i];
-                putativedegrees += lbcount * lwssb[i];
-                putativedegrees += macount * mwssa[i];
-                putativedegrees += mbcount * mwssb[i];
-                putativedegrees += gacount * glida[i];
-                putativedegrees += gbcount * glidb[i];
-                if (degrees[i] != putativedegrees) { return components; }
-            }
-
-            int hcount = 0;
-            int lcount = 0;
-            int mcount = 0;
-            int gcount = 0;
-
-            if (hacount >= 0 && hbcount >= 0) { hcount = hacount + hbcount; } else { return components; }
-            if (macount >= 0 && mbcount >= 0) { mcount = macount + mbcount; } else { return components; }
-            if (lacount >= 0 && lbcount >= 0) { lcount = lacount + lbcount; } else { return components; }
-            if (gacount >= 0 && gbcount >= 0) { gcount = gacount + gbcount; } else { return components; }
-
-            for (int i = 0; i < gcount; i++) { components.push_back("xq4_153"); }
-            for (int i = 0; i < lcount; i++) { components.push_back("xq4_6frc"); }
-            for (int i = 0; i < mcount; i++) { components.push_back("xq4_27dee6"); }
-            for (int i = 0; i < hcount; i++) { components.push_back("xq4_27deee6"); }
-
-            return components;
+            RETURN_SINGLETON();
 
         }
 
+
         std::vector<bitworld> getclusters(bitworld &live, bitworld &env, bool rigorous) {
 
+            (void) rigorous;
+
             bitworld lrem = live;
             std::vector<bitworld> clusters;
 
@@ -509,12 +393,7 @@ namespace apg {
                 bitworld cluster = grow_cluster(lrem.get1cell(), env, zoi);
                 cluster &= lrem;
                 lrem -= cluster;
-                if (rigorous) {
-                    pattern ppart(lab, lab->demorton(cluster, 1), rule);
-                    pseudoBangBang(ppart, &clusters, false);
-                } else {
-                    clusters.push_back(cluster);
-                }
+                clusters.push_back(cluster);
             }
 
             return clusters;
@@ -547,50 +426,24 @@ namespace apg {
             }
 
             repr = cl2.apgcode();
-            auto it = decompositions.find(repr);
-            if (it != decompositions.end()) {
-                elements = it->second;
-            } else {
+
+            if (repr[0] == 'x') {
+                elements = full_3dsep(repr);
+            } else if (recurse) {
                 uint64_t period = cl2.ascertain_period();
-                if ((M == 1) && (!b0)) {
-                    // 2-state rule:
-                    if ((zoi.length() <= 2) && (period == 4) && ((cl2.dx != 0) || (cl2.dy != 0))) {
-                        // Separating standard spaceships is considerably
-                        // faster and more reliable with sss; only fall
-                        // back on pbb if this fails:
-                        if ((repr == "xq4_27deee6zwsighhga") || (repr == "xq4_06eeed72zaghgis")) {
-                            // bypass sss()
-                            elements.push_back(repr);
-                        } else {
-                            elements = sss(cl2);
+                std::map<std::string, int64_t> rc = census(cl2, period << 3, 0, false);
+                for (auto it2 = rc.begin(); it2 != rc.end(); ++it2) {
+                    if (it2->second > 0) {
+                        for (int64_t i = 0; i < it2->second; i++) {
+                            elements.push_back(it2->first);
                         }
                     }
-                    if (elements.size() == 0) { elements = pseudoBangBang(cl2, 0, recurse); }
-                } else if (recurse && (!b0)) {
-                    // We may have a constellation since we've bypassed PseudoBangBang.
-                    // This tends to be problematic in B0 and Generations rules with
-                    // infinite-growth patterns, so we try to reseparate the pattern:
-                    uint64_t period = cl2.ascertain_period();
-                    std::map<std::string, int64_t> rc = census(cl2, period << 3, 0, false);
-                    for (auto it2 = rc.begin(); it2 != rc.end(); ++it2) {
-                        if (it2->second > 0) {
-                            for (int64_t i = 0; i < it2->second; i++) {
-                                elements.push_back(it2->first);
-                            }
-                        }
-                    }
-                } else {
-                    // PseudoBangBang not supported for multistate rules; skip separation:
-                    elements.push_back(repr);
                 }
-
-                // better separation:
-                elements = map_3dsep(elements);
-
-                decompositions[repr] = elements;
+            } else {
+                elements.push_back(repr);
             }
 
-            if (bb != 0) {
+            if ((repr[0] == 'x') && (bb != 0)) {
                 bitcache.emplace(bb, repr);
             }
             
